import inspect
import logging
import time
import weakref
from functools import partial
from collections import defaultdict


logger = logging.getLogger(__name__)


class Tag(object):
    """``class`` that abstract tags and act as a simple container."""
    def __init__(self, name, valuable, ttl=-1, refresh=None, parent=None):
        """
        :param string name: tag name
        :param valuable: something that gives tag value, string or callable
        :param None,int ttl: Time to live for caching the tags, possible values
            are:
                * None (means no ttl)
                * -1 (means infinite ttl)
                * positive integer in seconds

            Note: this has absolutely not any importance for the cc-node, only
                for cc-server
        :param None,int refresh: period used to refresh tag value on the node
        :param object obj: parent object the tag is attached to, it may be
            needed for the tag callable to process the tag
        """
        self.name = name
        self.value = None
        self.is_function = False
        self.ttl = ttl
        self.refresh = refresh
        self.parent = parent if parent is None else weakref.proxy(parent)


        if inspect.isfunction(valuable):
            self.is_function = True
            args_count = len(inspect.getargspec(valuable).args)
            if args_count > 1:
                raise TypeError('Tag function take at most 1 argument')
            elif args_count == 1:
                self._calculate_value = partial(valuable, self.parent)
            else:
                self._calculate_value = valuable
        else:
            self.value = valuable

        self.watcher = None

        # FIXME special arguments for tag db
        self.db = None
        self.sub_id = None

    def calculate_value(self):
        self.value = self._calculate_value()
        # logger.debug('Calculate Tag(%s) = %s', self.name, self.value)

    def start(self, loop):
        """
        :param loop: pyev loop
        """
        if not self.is_function:
            return

        if self.refresh is None:
            self.calculate_value()
            return

        # TODO more sophisticated calculation with event propagation
        self.watcher = loop.timer(.0, float(self.refresh), lambda *args:
                                  self.calculate_value())
        self.watcher.start()

    def stop(self):
        if self.watcher is not None:
            self.watcher.stop()


def tag_inspector(mod, parent=None):
    """Inspect module to find tags.

    :param module mod: module to inspect

    Currently there are two ways to define a tag inside a module:

        * affect a string to a variable, the variable name will be the tag
          name
        * define a function that returns a value as a string or None (meaning
          the tag doesn't exist on the host), as you guessed the function name
          will define the tag name
    """
    tags = []
    for n, m in inspect.getmembers(mod):  # (name, member)
        # only keep strings or functions as tags
        if getattr(m, '__module__', None) != mod.__name__ or (
            n.startswith('_')):
            continue
        elif isinstance(m, (str, unicode)):
            # if string, it means it is constant, then ttl = -1
            ttl = -1
            refresh = None
        elif inspect.isfunction(m):
            # if function take function ttl argument or set -1 by default
            ttl = getattr(m, 'ttl', -1)
            refresh = getattr(m, 'refresh', None)
        else:
            # whatever it is we don't care...
            continue

        logger.debug('Introspected %s with ttl %s.', n, ttl)

        # finally add the tag
        tags.append(Tag(n, m, ttl, refresh, parent))

    return tags


# decorators for tag inspector
def ttl(value):
    def decorator(func):
        func.ttl = value
        return func
    return decorator


def refresh(value):
    def decorator(func):
        func.refresh = value
        return func
    return decorator


def get_tags(tags_dict, tags=None, noresolve_tags=None):
    """Helper to get tags.

    :param tags_dict: dict containing :class:`Tag` objects
    :param tags: list of tags to get (None mean all tags)
    :param noresolve_tags: list of tags to exclude from result
    """
    logger.debug('Tags request: %s, %s', unicode(tags), unicode(noresolve_tags))

    tags = set(tags) - set(noresolve_tags) if tags is not None else None

    if tags is None:
        tags = tags_dict.iterkeys()
    else:
        tags = tags & set(tags_dict)

    profile = time.time()
    result = dict((
        t,  # tag name
        dict(
            value=tags_dict[t].value,
            ttl=tags_dict[t].ttl,
        ),
    ) for t in tags)

    logger.debug('Profiling: %f seconds', time.time() - profile)
    logger.debug('Returning: %s', unicode(result))
    return result


class TagDB(object):
    """Tag database. FIXME comment

    Handles common operations such as registering tag on the cc-server, updating
    its values, etc.

    TagDB can have a parent TagDB, in this case, the latter will handle tag
    registration on the cc-server.

    """
    def __init__(self, main, parent_db=None, tags=None, sub_tags=None):
        """
        :param main: MainLoop instance
        :param TagDB parent_db: TagDB parent object
        :param iterable tags: initial tags
        :param dict sub_tags: initial subtags
        """
        self.main = weakref.proxy(main)
        self.parent = parent_db

        if tags is None:
            tags = tuple()
        if sub_tags is None:
            sub_tags = {}

        self.db = defaultdict(
            dict,
            __main__=dict(),  # tags for main object
            # others objects
        )

        for tag in tags:
            self.add_tag(tag)

        for sub_id, tags in sub_tags.iteritems():
            for tag in tags:
                self.add_sub_tag(sub_id, tag)

        #: dict for async call storage, keep a part of log message
        self.async_calls = dict()

    # RPC part
    def rpc_register(self):
        """Register all objects and tags on the cc-server.

        This is used just after a (re)connection to the cc-server is made.
        """
        for sub_id in self.db:
            if sub_id == '__main__':
                continue
            self.rpc_register_sub_object(sub_id)

        # TODO register tags

    def rpc_register_sub_object(self, sub_id):
        # register object on the cc-server
        if self.main.rpc_authenticated:
            self.async_calls[self.main.rpc_con.rpc.async_call_cb(
                self.rpc_object_register_cb,
                'register',
                sub_id,
                'vm',  # FIXME generalization
            )] = sub_id

    def rpc_unregister_sub_object(self, sub_id):
        if self.main.rpc_authenticated:
            self.async_calls[self.main.rpc_con.rpc.async_call_cb(
                self.rpc_object_unregister_cb,
                'unregister',
                sub_id,
            )] = sub_id

    def rpc_object_register_cb(self, call_id, response, error):
        name = self.async_calls.pop(call_id)
        if error is not None:
            logger.error('Error while trying to register the object %s,'
                         ' %s("%s")', name, error['exception'],
                         error['message'])
            return

        logger.info('Successfully registered object %s', name)

    def rpc_object_unregister_cb(self, call_id, response, error):
        name = self.async_calls.pop(call_id)
        if error is not None:
            logger.error('Error while trying to unregister the object %s,'
                         ' %s("%s")', name, error['exception'],
                         error['message'])
            return

        logger.info('Successfully unregistered object %s', name)

    def rpc_tag_register_cb(self, call_id, response, error):
        pass

    def rpc_tag_unregister_cb(self, call_id, response, error):
        pass

    def rpc_sub_tag_register_cb(self, call_id, response, error):
        pass

    def rpc_sub_tag_unregister_cb(self, call_id, response, error):
        pass
    # end RPC part

    # tag handling part, used by plugins
    def add_tags(self, tags):
        """
        :param iterable tags: list of tags to add
        """
        for tag in tags:
            self.add_tag(tag)

    def add_sub_tags(self, sub_id, tags):
        for tag in tags:
            self.add_sub_tag(sub_id, tag)

    def remove_tags(self, tag_names):
        """
        :param iterable tag_names: list of tag names to remove
        """
        for name in tag_names:
            self.remove_tag(name)

    def add_tag(self, tag):
        # set special attributes on tag instance
        if self.parent is None:
            tag.db = self
            tag.sub_id = '__main__'
            tag.start(self.main.evloop)
            # TODO register tag
        else:
            self.parent.add_tag(tag)
        self.db['__main__'][tag.name] = tag

    def remove_tag(self, tag_name):
        tag = self.db['__main__'].pop(tag_name)
        if self.parent is None:
            tag.db = None
            tag.sub_id = None
            tag.stop()
            # TODO unregister tag
        else:
            self.parent.remove_tag(tag_name)

    def add_sub_tag(self, sub_id, tag):
        if self.parent is None:
            tag.db = self
            tag.sub_id = sub_id
            tag.start(self.main.evloop)
            # TODO register tag
        else:
            self.parent.add_sub_tag(sub_id, tag)
        self.db[sub_id][tag.name] = tag

    def remove_sub_tag(self, sub_id, tag_name):
        tag = self.db[sub_id].pop(tag_name)
        if self.parent is None:
            tag.db = None
            tag.sub_id = None
            tag.stop()
            # TODO unregister tag
        else:
            self.parent.remove_sub_tag(sub_id, tag_name)

    def add_sub_object(self, sub_id, tags):
        if self.parent is None:
            self.rpc_register_sub_object(sub_id)
        else:
            # tags will be registered after
            self.parent.add_sub_object(sub_id, tuple())
        # register tags
        for t in tags:
            self.add_sub_tag(sub_id, t)

    def remove_sub_object(self, sub_id):
        if self.parent is None:
            for name in self.db[sub_id].keys():
                self.remove_sub_tag(sub_id, name)
            del self.db[sub_id]
            self.rpc_unregister_sub_object(sub_id)
        else:
            sub_tags = self.db.pop(sub_id)
            self.parent.remove_sub_object(sub_id)

    def update_tag_value(self, tag):
        """Update tag value on cc-server."""
        # TODO

    # dict like
    def get(self, key, default=None):
        return self.db.get(key, default)

    def __getitem__(self, key):
        return self.db[key]

    def keys(self):
        return self.db.keys()

    def iteritems(self):
        return self.db.iteritems()

    def itervalues(self):
        return self.db.itervalues()

    def update_from_db(self, tag_db):
        """Update self tag db from other instance."""
        self.add_tags(tag_db['__main__'].itervalues())

        for sub_id, db in tag_db.iteritems():
            if sub_id == '__main__':
                continue
            self.add_sub_tags(sub_id, db.itervalues())

    def delete_from_db(self, tag_db):
        self.remove_tags(tag_db['__main__'])
        for sub_id in tag_db:
            if sub_id == '__main__':
                continue
            self.remove_suobject(sub_id)
